/*******************************************************

	Created by : Joe Forte
	File : ParticleSystem.cpp
	Desc : implements the ParticleSystem class.

********************************************************/

#include <cmath>
#include <algorithm>
#include <assert.h>
#include "ParticleSystem.h"

const float PI = 3.14159265f;
const float PI_2 = PI / 2.0f;
const float PI_4 = PI / 4.0f;

struct Vertex
{
	float x, y, z;
	float r, g, b, a;
	float u, v;
};

const size_t VERTEX_SIZE = sizeof(Vertex);
const size_t VERTEX_COLOR_OFFSET = 3 * sizeof(float);
const size_t VERTEX_TEX_OFFSET = 7 * sizeof(float);

//------------------------------------------------------
bool ParticleCmp(const ParticleSystem::SimpleParticle &p1, const ParticleSystem::SimpleParticle &p2)
{
	return p1.transP.z < p2.transP.z;
}

//------------------------------------------------------
float RandFloat(float min, float max)
{
	return (max - min)*((float)rand()/(float)RAND_MAX) + min;
}

//------------------------------------------------------
ParticleSystem::ParticleSystem() : initFunc(NULL), alphaFunc(AlphaFuncConst), sizeFunc(NULL), colorFunc(NULL)
{
	vertexBuffer = 0;
	colorBuffer = 0;
	texCoordBuffer = 0;

	timer = 0.0f;
	spawned = 0;
}

//------------------------------------------------------
ParticleSystem::~ParticleSystem()
{
	glDeleteBuffersARB(1, &vertexBuffer);
	glDeleteBuffersARB(1, &colorBuffer);
	glDeleteBuffersARB(1, &texCoordBuffer);
}

// Windows header stupidly defines a max macro making std::max impossible to use
#ifdef _WIN32
	#undef max
	#undef min
#endif

//------------------------------------------------------
void ParticleSystem::Update(float dt)
{
	// Determines the number of particles to spawn
	timer = timer + dt;
	int spawn = std::max(0, std::min((int)std::floor(timer * spawnRate) - spawned, (int)spawnRate));
	spawned = spawned + spawn;

	if(timer >= 1.0f)
	{
		timer = 0.0f;
		spawned = 0;
	}
	
	SpawnParticles(spawn);

	const Vector3f a = g + wind;
	const Vector3f at = a * dt;
	const Vector3f att = at * dt;

	std::vector<size_t>::iterator i = active.begin();
	while(i != active.end())
	{
		size_t index = *i;

		Particle &p = particleStore[ index ];

		p.age += dt;
		
		p.alpha = alphaFunc(p.age, particleLife, p.alpha);

		if(sizeFunc)
			sizeFunc(p.w, p.h, p.age, particleLife);

		if(colorFunc)
			colorFunc(p.r, p.g, p.b, p.age, particleLife);

		if( p.age > particleLife )
		{
			i = active.erase(i);
			dead.push_back(index);
		}
		else
		{
			for(std::vector<Plane>::iterator pl = planes.begin(); pl != planes.end(); pl ++)
			{
				Plane &plane = *pl;

				// Computes distance to the plane
				float d = Dot(plane.n, p.p) + plane.d;
				if( d < 0.0f )
				{
					// Corrects any penetration
					p.p = p.p + -d*plane.n;

					p.v = bounce*p.v.Reflect(plane.n);
				}
			}

			p.p = p.p + p.v*dt + att;
			p.v = p.v + at;
			i++;
		}
	}
}

//------------------------------------------------------
void ParticleSystem::Draw(Camera &cam, bool useTex)
{
	if(useTex)
		glBindTexture(GL_TEXTURE_2D, texID);

	glMatrixMode(GL_MODELVIEW);
	glPushMatrix();
	glLoadIdentity();

	Matrixf trans = cam.GetViewMatrix();

	size_t len = active.size();

	std::vector<SimpleParticle> toDraw;

	for(size_t i = 0; i < len; i++)
	{
		SimpleParticle sp;

		sp.transP = trans.Transform( particleStore[ active[i] ].p );
		sp.index = active[i];

		toDraw.push_back(sp);
	}

	std::sort(toDraw.begin(), toDraw.end(), ParticleCmp);

#ifdef USE_VBO

	glBindBufferARB(GL_ARRAY_BUFFER_ARB, vertexBuffer);
	//glBufferDataARB(GL_ARRAY_BUFFER_ARB, 3 * 4 * maxParticles * sizeof(float), NULL, GL_STREAM_DRAW_ARB);
	float *vertexPtr = (float*)glMapBufferARB(GL_ARRAY_BUFFER_ARB, GL_WRITE_ONLY_ARB);

	assert(vertexPtr);

	for(size_t i = 0, count = 0; count < len; count++)
	{
		Particle &prt = particleStore[ toDraw[count].index ];
		Point3f &p = toDraw[count].transP;
		
		float w = prt.w / 2.0f;
		float h = prt.h / 2.0f;

		vertexPtr[i++] = p.x - w; vertexPtr[i++] = p.y - h; vertexPtr[i++] = p.z;
		vertexPtr[i++] = p.x + w; vertexPtr[i++] = p.y - h; vertexPtr[i++] = p.z;
		vertexPtr[i++] = p.x + w; vertexPtr[i++] = p.y + h; vertexPtr[i++] = p.z;
		vertexPtr[i++] = p.x - w; vertexPtr[i++] = p.y + h; vertexPtr[i++] = p.z;
	}

	glUnmapBufferARB(GL_ARRAY_BUFFER_ARB);

	glBindBufferARB(GL_ARRAY_BUFFER_ARB, colorBuffer);
	//glBufferDataARB(GL_ARRAY_BUFFER_ARB, 4 * 4 * maxParticles * sizeof(float), NULL, GL_STREAM_DRAW_ARB);
	float *colorPtr = (float*)glMapBufferARB(GL_ARRAY_BUFFER_ARB, GL_WRITE_ONLY_ARB);

	assert(colorBuffer);
	
	for(size_t i = 0, count = 0; count < len; count++)
	{
		Particle &prt = particleStore[ toDraw[count].index ];

		//       r                    g                       b                    a
		colorPtr[i++] = 1.0f; colorPtr[i++] = 1.0f; colorPtr[i++] = 1.0f; colorPtr[i++] = prt.alpha;
		colorPtr[i++] = 1.0f; colorPtr[i++] = 1.0f; colorPtr[i++] = 1.0f; colorPtr[i++] = prt.alpha;
		colorPtr[i++] = 1.0f; colorPtr[i++] = 1.0f; colorPtr[i++] = 1.0f; colorPtr[i++] = prt.alpha;
		colorPtr[i++] = 1.0f; colorPtr[i++] = 1.0f; colorPtr[i++] = 1.0f; colorPtr[i++] = prt.alpha;
	}

	glUnmapBufferARB(GL_ARRAY_BUFFER_ARB);

	// Draws buffered data
	glBindBufferARB(GL_ARRAY_BUFFER_ARB, vertexBuffer);
	glEnableClientState(GL_VERTEX_ARRAY);
	glVertexPointer(3, GL_FLOAT, 0, 0);

	glBindBufferARB(GL_ARRAY_BUFFER_ARB, colorBuffer);
	glEnableClientState(GL_COLOR_ARRAY);
	glColorPointer(4, GL_FLOAT, 0, 0);

	glBindBufferARB(GL_ARRAY_BUFFER_ARB, texCoordBuffer);
	glEnableClientState(GL_TEXTURE_COORD_ARRAY);
	glTexCoordPointer(2, GL_FLOAT, 0, 0);

	glDrawArrays(GL_QUADS, 0, (GLsizei)len);

	glDisableClientState(GL_VERTEX_ARRAY);
	glDisableClientState(GL_COLOR_ARRAY);
	glDisableClientState(GL_TEXTURE_COORD_ARRAY);
	glBindBufferARB(GL_ARRAY_BUFFER_ARB, 0);

#else

	glBegin(GL_QUADS);

	for(size_t i = 0; i < len; i++)
	{
		Particle &prt = particleStore[ toDraw[i].index ];
		Point3f &p = toDraw[i].transP;

		float w = prt.w / 2.0f;
		float h = prt.h / 2.0f;

		glColor4f(prt.r, prt.g, prt.b, prt.alpha);

		glTexCoord2f(0.0f, 0.0f);
		glVertex3f(p.x - w, p.y - h, p.z);

		glTexCoord2f(1.0f, 0.0f);
		glVertex3f(p.x + w, p.y - h, p.z);

		glTexCoord2f(1.0f, 1.0f);
		glVertex3f(p.x + w, p.y + h, p.z);

		glTexCoord2f(0.0f, 1.0f);
		glVertex3f(p.x - w, p.y + h, p.z);
	}
	
	glColor4f(1.0f, 1.0f, 1.0f, 1.0f);

	glEnd();

#endif

	glPopMatrix();

	glBindTexture(GL_TEXTURE_2D, 0);
}

//------------------------------------------------------
void ParticleSystem::SpawnParticles(int amt)
{
	while(amt > 0 && active.size() < (size_t)maxParticles)
	{
		size_t index = dead.back();
		dead.pop_back();

		if(initFunc)
			initFunc(*this, particleStore[index]);
		else
			InitParticle(index);

		active.push_back(index);

		amt--;
	}
}

//------------------------------------------------------
void ParticleSystem::Reset()
{
	active.clear();
	dead.clear();
	
	for(size_t i = 0; i < (size_t)maxParticles; i++)
	{
		dead.push_back(i);
	}
}

//------------------------------------------------------
void ParticleSystem::InitParticle(size_t i)
{
	const float vel = 10.0f;
	const float x = RandFloat(-1.0f, 1.0f);
	const float z = RandFloat(-1.0f, 1.0f);

	Particle &p = particleStore[i];
	p.p = startLocation;
	p.v = Vector3f(2.0f*x, vel, 2.0f*z);
	p.w = 1.5f;
	p.h = 1.5f;
	p.r = 1.0f;
	p.g = 1.0f;
	p.b = 1.0f;
	p.alpha = 1.0f;
	p.age = 0.0f;
}

//------------------------------------------------------
void ParticleSystem::Init()
{
	particleStore.reserve( maxParticles );
	dead.reserve(maxParticles);

	for(size_t i = 0; i < (size_t)maxParticles; i++)
	{
		Particle p;

		particleStore.push_back(p);
		dead.push_back(i);
	}
#ifdef USE_VBO
	InitBuffers();
#endif
}

//------------------------------------------------------
void ParticleSystem::InitBuffers()
{
	size_t vertexSize = 3 * 4 * maxParticles * sizeof(float);
	size_t colorSize = 4 * 4 * maxParticles * sizeof(float);
	size_t texCoordSize = 2 * 4 * maxParticles * sizeof(float);

	if(!vertexBuffer)
	{
		glDeleteBuffersARB(1, &vertexBuffer);
		glDeleteBuffersARB(1, &colorBuffer);
		glDeleteBuffersARB(1, &texCoordBuffer);
	}

	glGenBuffersARB(1, &vertexBuffer);
	glGenBuffersARB(1, &colorBuffer);
	glGenBuffersARB(1, &texCoordBuffer);

	glBindBufferARB(GL_ARRAY_BUFFER_ARB, vertexBuffer);
	glBufferDataARB(GL_ARRAY_BUFFER_ARB, vertexSize, NULL, GL_STREAM_DRAW_ARB);

	glBindBufferARB(GL_ARRAY_BUFFER_ARB, colorBuffer);
	glBufferDataARB(GL_ARRAY_BUFFER_ARB, colorSize, NULL, GL_STREAM_DRAW_ARB);


	// Creates the static texture data
	size_t len = 2 * 4 * maxParticles;
	float *texCoords = new float[len];

	{
		size_t i = 0;
		while(i < len)
		{
			//      u                       v 
			texCoords[i++] = 0.0f;  texCoords[i++] = 0.0f;
			texCoords[i++] = 0.0f;  texCoords[i++] = 1.0f;
			texCoords[i++] = 1.0f;  texCoords[i++] = 1.0f;
			texCoords[i++] = 1.0f;  texCoords[i++] = 0.0f;
		}
	}

	glBindBufferARB(GL_ARRAY_BUFFER_ARB, texCoordBuffer);
	glBufferDataARB(GL_ARRAY_BUFFER_ARB, texCoordSize, (void*)texCoords, GL_STATIC_DRAW_ARB);
	
	delete texCoords;

	glBindBufferARB(GL_ARRAY_BUFFER_ARB, 0);
}

//------------------------------------------------------
void SmokeCloudParticleInit(ParticleSystem &psys, ParticleSystem::Particle &p)
{
	const float x = RandFloat(-1.0f, 1.0f);
	const float y = RandFloat(-1.0f, 1.0f);
	const float z = RandFloat(-1.0f, 1.0f);
	
	p.p = psys.GetStartLocation();
	p.v = Vector3f(x, y, z);
	p.w = 5.0f;
	p.h = 5.0f;
	p.r = 1.0f;
	p.g = 1.0f;
	p.b = 1.0f;
	p.alpha = 1.0f;
	p.age = 0.0f;
}

//------------------------------------------------------
void SmokeStackParticleInit(ParticleSystem &psys, ParticleSystem::Particle &p)
{
	const float x = RandFloat(-1.0f, 1.0f);
	const float y = RandFloat( 0.0f, 1.0f);
	const float z = RandFloat(-1.0f, 1.0f);
	
	p.p = psys.GetStartLocation();
	p.v = Vector3f(x, y, z);
	p.w = 5.0f;
	p.h = 5.0f;
	p.r = 1.0f;
	p.g = 1.0f;
	p.b = 1.0f;
	p.alpha = 1.0f;
	p.age = 0.0f;
}

//------------------------------------------------------
float AlphaFuncConst(float age, float lifeTime, float currAlpha)
{
	return currAlpha;
}

//------------------------------------------------------
float AlphaFuncLinear(float age, float lifeTime, float currAlpha)
{
	float t = age / lifeTime;
	return 1.0f - t;
}

//------------------------------------------------------
float AlphaFuncPoly2(float age, float lifeTime, float currAlpha)
{
	float t = age / lifeTime;
	return 1.0f - t*t;
}

//------------------------------------------------------
float AlphaFuncPoly3(float age, float lifeTime, float currAlpha)
{
	float t = age / lifeTime;
	return 1.0f - t*t*t;
}

//------------------------------------------------------
float AlphaFuncPoly4(float age, float lifeTime, float currAlpha)
{
	float t = age / lifeTime;
	return 1.0f - t*t*t*t;
}

//------------------------------------------------------
void SmokeStackSizeFunc(float &w, float &h, float age, float lifeTime)
{
	w = 5.0f * (age / lifeTime) + 5.0f;
	h = w;
}

//------------------------------------------------------
void SmokeStackColorFunc(float &r, float &g, float &b, float age, float lifeTime)
{
	float t = age / lifeTime;
	r = g = b = t + 0.3f;
}
