/*******************************************************

	Created by : Joe Forte
	File : ParticleSystem.h
	Desc : defines the ParticleSystem class.

********************************************************/

#ifndef _PARTICLESYSTEM_H
#define _PARTICLESYSTEM_H

#include <GL/glew.h>

#ifdef _WIN32
	#include "windows.h"
#endif

#ifdef __APPLE__
	#include <OpenGL/gl.h>
#else
	#include <GL/gl.h>
#endif

#include <vector>

#include "Vector3.h"
#include "Point3.h"
#include "Camera.h"

struct Plane
{
	Vector3f n;
	float d;
};

class ParticleSystem
{
public:
	ParticleSystem();
	~ParticleSystem();

	void Init(); // call after the system has been set up

	void SetGravity(const Vector3f &_g) { g = _g; }
	void SetWind(const Vector3f &w) { wind = w; }
	void SetStartPosition(const Point3f &p) { startLocation = p; }
	void SetMaxParticles(int max) { maxParticles = max; }
	void SetTexture(GLuint id) { texID = id; }
	void SetParticleLife(float l) { particleLife = l; }
	void SetSpawnRate(float r) { spawnRate = r; }
	void SetBounce(float b) { bounce = b; }
	void AddPlane(const Plane &p) { planes.push_back(p); }

	void SpawnParticles(int amt);

	void IncSpawnRate(float amt) { spawnRate += amt; }

	void Reset();

	size_t ParticleCount() const { return active.size(); }
	Vector3f GetWind() { return wind; }
	float GetSpawnRate() { return spawnRate; }

	void Update(float dt);

	void Draw(Camera &cam, bool useTex);

	Point3f GetStartLocation() { return startLocation; }
	GLuint TextureID() { return texID; };

	struct Particle
	{
		Vector3f v;
		Point3f p;

		float w;
		float h;
		float r, g, b, alpha;
		float age;
	};

	struct SimpleParticle
	{
		Point3f transP;
		size_t index;
	};

	typedef void (*fnParticleInit)(ParticleSystem &psys, ParticleSystem::Particle &p);

	void SetInitFunc(fnParticleInit func) { initFunc = func; }

	typedef float (*fnAlphaFunc)(float age, float lifeTime, float currAlpha);

	void SetAlphaFunc(fnAlphaFunc func) { if(func) alphaFunc = func; }

	typedef void (*fnSizeFunc)(float &w, float &h, float age, float lifeTime);

	void SetSizeFunc(fnSizeFunc func) { sizeFunc = func; }

	typedef void (*fnColorFunc)(float &r, float &g, float &b, float age, float lifeTime);

	void SetColorFunc(fnColorFunc func) { colorFunc = func; }

private:

	void InitBuffers();

	void InitParticle(size_t i);

	fnParticleInit initFunc;
	fnAlphaFunc alphaFunc;
	fnSizeFunc sizeFunc;
	fnColorFunc colorFunc;

	// Acceleration vectors for gravity and wind
	Vector3f g;
	Vector3f wind;

	std::vector<Plane> planes;
	float bounce;
	Point3f startLocation;
	float spawnRate;
	float particleLife;
	float timer;
	int spawned;

	std::vector<size_t> active; // stores indices into the particleStore
	std::vector<size_t> dead;

	std::vector<Particle> particleStore;

	int maxParticles;

	GLuint texID;
	GLuint vertexBuffer;
	GLuint colorBuffer;
	GLuint texCoordBuffer;
	GLuint indexBuffer;
};

void SmokeCloudParticleInit(ParticleSystem &psys, ParticleSystem::Particle &p);
void SmokeStackParticleInit(ParticleSystem &psys, ParticleSystem::Particle &p);

// Alpha functions
float AlphaFuncConst(float age, float lifeTime, float currAlpha);
float AlphaFuncLinear(float age, float lifeTime, float currAlpha);
float AlphaFuncPoly2(float age, float lifeTime, float currAlpha);
float AlphaFuncPoly3(float age, float lifeTime, float currAlpha);
float AlphaFuncPoly4(float age, float lifeTime, float currAlpha);

// Size functions
void SmokeStackSizeFunc(float &w, float &h, float age, float lifeTime);

// Color functions
void SmokeStackColorFunc(float &r, float &g, float &b, float age, float lifeTime);

#endif
