/*******************************************************

	Created by : Joe Forte
	File : main.cpp
	Desc : defines the entry point for the Soft
	Particle test program.

********************************************************/

#ifdef _WIN32
	#include "windows.h"
#endif

#include <GL/glew.h>

#ifdef __APPLE__
	#include <OpenGL/gl.h>
	#include <GLUT/glut.h>
#else
	#include <GL/gl.h>
	#include <GL/glut.h>
#endif

#include <iostream>
#include <fstream>
#include <sstream>
#include <string>
#include <ctime>
#include <limits>
#include <float.h>

#include "SOIL.h" // Uses SOIL ('Simple OpenGL Image Loader') to load texture data 

#include "Vector3.h"
#include "Color4.h"
#include "Camera.h"
#include "ParticleSystem.h"

// Useful constants
const float PI = 3.14159265f;
const float PI_2 = PI / 2.0f;
const int MAX_FPS = 60;
const float MAX_FRAME_TIME_MS = 1000.0f / MAX_FPS;
const int DEFAULT_WIDTH = 800;
const int DEFAULT_HEIGHT = 600;
const Color4f bgColor(0.0f, 0.0f, 0.0f);

// App data
const std::string WINDOW_TITLE = "Soft Particles";
bool pausePSystem = false;
bool softParticleSupport = true;
bool useSoftParticles = false;
bool useTerrain = false;
bool mouseZoom = false;
bool mouseMove = false;
bool showInfo = true;
int softParticlePower = 2;
int lastMouseX;
int lastMouseY;
int WIDTH = DEFAULT_WIDTH;
int HEIGHT = DEFAULT_HEIGHT;

// Cam data
const Point3f CAM_EYEPOINT(8.0f, 8.0f, 6.0f);
const Point3f CAM_LOOKAT(0.0f, 0.0f, 0.0f);
const Vector3f CAM_UP(0.0f, 1.0f, 0.0f);
Matrixf gProj;
Camera gCam;

// FPS data
int frames = 0;
float fps = 0.0f;
float timer = 0.0f;

// texture data
const std::string FLOOR_CENTER_IMAGE_FILE = "./Data/grid_center.png";
const std::string FLOOR_EDGE_IMAGE_FILE = "./Data/grid_edge.png";
const std::string PARTICLE_FOUNTAIN_IMAGE_FILE = "./Data/test_particle.tga";
const std::string PARTICLE_SMOKE_IMAGE_FILE = "./Data/smoke_particle.dds";

GLuint floorCenterTexID = 0;
GLuint floorEdgeTexID	= 0;
GLuint particleTexID	= 0;
GLuint smokeTexID		= 0;

// Particle system data
const Vector3f WIND(-3.0f, 1.5f, -5.0f);
const float SPAWN_RATE_INC = 10.0f;
ParticleSystem gFountain;
ParticleSystem gSmokeCloud;
ParticleSystem gSmokeStack;
ParticleSystem *pPSystem = NULL;

// Timing data 
int gPrevTime;
float dt;

// Shader and frame buffer data
GLhandleARB fbo				= 0;
GLhandleARB depthTex		= 0;
GLhandleARB depthShader		= 0;
GLhandleARB depthVS			= 0;
GLhandleARB depthPS			= 0;
GLhandleARB particleShader	= 0;
GLhandleARB particleVS		= 0;
GLhandleARB particlePS		= 0;

#ifdef _WIN32
	const std::string DEPTH_OUT_VS_FILE = ".\\Data\\depthOut.vs";
	const std::string DEPTH_OUT_PS_FILE = ".\\Data\\depthOut.ps";
	const std::string PARTICLE_SYSTEM_VS_FILE = ".\\Data\\particleSystem.vs";
	const std::string PARTICLE_SYSTEM_PS_FILE = ".\\Data\\particleSystem.ps";
#else
	const std::string DEPTH_OUT_VS_FILE = "./Data/depthOut.vs";
	const std::string DEPTH_OUT_PS_FILE = "./Data/depthOut.ps";
	const std::string PARTICLE_SYSTEM_VS_FILE = "./Data/particleSystem.vs";
	const std::string PARTICLE_SYSTEM_PS_FILE = "./Data/particleSystem.ps";
#endif

bool InitProgram(int w, int h);
bool ReInitProgram(int w, int h);
bool InitFBO();
bool InitShaders();
void RenderScene();
void DrawFPS();
void DrawSceneInfo();
void TakeScreenShot();

void CleanUp();

void glutDisplayCallback();
void glutIdleFuncCallback();
void glutWindowReshapeCallback(int w, int h);
void glutKeyboardCallback(unsigned char key, int x, int y);
void glutMouseCallback(int button, int state, int x, int y);
void glutMouseMoveCallback(int x, int y);

void glutDrawStr(const Color4f &c, float x, float y, const char *str);

int main(int argc, char *argv[])
{
	// Sets up glut
	glutInit(&argc, argv);
	glutInitWindowSize(WIDTH, HEIGHT);
	glutInitDisplayMode(GLUT_RGBA | GLUT_DOUBLE | GLUT_DEPTH);
	glutCreateWindow( WINDOW_TITLE.c_str() );

	// Sets up glew
	GLenum err = glewInit();
	if(err != GLEW_OK)
	{
		std::cerr << "ERROR: GLEW could not be initialized " << glewGetErrorString(err) << std::endl;
		return 1;
	}

	if( !GLEW_ARB_shader_objects || !GLEW_EXT_framebuffer_object || !GLEW_ARB_texture_float )
	{
		std::cerr << "ERROR: The needed OpenGL extenstions are not supported on this machine.\n";
		std::cerr << "       The following extenstions are needed:\n";
		std::cerr << "           GL_ARB_shader_objects\n";
		std::cerr << "           GL_EXT_framebuffer_object\n";
		std::cerr << "           GL_ARB_texture_float\n";
		std::cerr << "           GL_ARB_color_buffer_float\n";

#ifdef _WIN32
		system("pause");
#endif
		return 1;
	}

	if(!GLEW_ARB_color_buffer_float)
	{
		std::cerr << "ERROR: Soft particles can not be used on this machine.  The needed hardware support is not availible.\n";
		softParticleSupport = false;
	}

	// Sets callbacks
	glutDisplayFunc( glutDisplayCallback );
	glutIdleFunc( glutIdleFuncCallback );
	glutReshapeFunc( glutWindowReshapeCallback );
	glutKeyboardFunc( glutKeyboardCallback );
	glutMouseFunc( glutMouseCallback );
	glutMotionFunc( glutMouseMoveCallback );

	// Sets up program data
	if( !InitProgram(WIDTH, HEIGHT) )
	{
		std::cerr << "Program Data could not be initialized!!!\n";
		return 1;
	}

	gPrevTime = glutGet(GLUT_ELAPSED_TIME);
	dt = 0.0f;

	// Starts the never ending loop...
	glutMainLoop();

	return 0;
}

//------------------------------------------------------
void glutDisplayCallback()
{
	frames++;
	timer += dt;
	if( timer >= 1.0f )
	{
		fps = frames / timer;
		timer = 0.0f;
		frames = 0;
	}

	RenderScene();
}

//------------------------------------------------------
void glutIdleFuncCallback()
{
	int curTime = glutGet(GLUT_ELAPSED_TIME);
	
	// Fixes the frame rate
	float tm = curTime - gPrevTime;
	if( tm < MAX_FRAME_TIME_MS )
		return;
	
	dt = tm / 1000.0f;

	if(!pausePSystem)
		pPSystem->Update(dt);

	gPrevTime = curTime;

	glutPostRedisplay();
}

//------------------------------------------------------
void glutWindowReshapeCallback(int w, int h)
{
	if( !ReInitProgram(w, h) )
	{
		std::cerr << "Program data could not be reinitialized on window reshape!!\n";
	}
}

//------------------------------------------------------
void glutKeyboardCallback(unsigned char key, int x, int y)
{
	switch(key)
	{
	case 27: // 'ESC' key
		CleanUp();
		exit(0);
		break;
	case 'w':
		gCam.RotateUp( -PI * dt );
		break;
	case 's':
		gCam.RotateUp( PI * dt );
		break;
	case 'a':
		gCam.RotateRight( -PI * dt );
		break;
	case 'd':
		gCam.RotateRight( PI * dt );
		break;
	case 'z':
		gCam.Zoom( 1.0f );
		break;
	case 'x':
		gCam.Zoom( -1.0f );
		break;
	case '+':
		pPSystem->IncSpawnRate(SPAWN_RATE_INC);
		break;
	case '-':
		pPSystem->IncSpawnRate(-SPAWN_RATE_INC);
		break;
	case '<':
	case ',':
		softParticlePower--;
		if(softParticlePower < 1) softParticlePower = 1;
		break;
	case '>':
	case '.':
		softParticlePower++;
		break;
	case 'v':
		if(pPSystem->GetWind() == WIND)
			pPSystem->SetWind(Vector3f(0.0f, 0.0f, 0.0f));
		else
			pPSystem->SetWind(WIND);
		break;
	case 'f':
		if(softParticleSupport)
			useSoftParticles = !useSoftParticles;
		break;
	case 'p':
		pausePSystem = !pausePSystem;
		break;
	case 'r': // Reset camera
		gCam.Set(CAM_EYEPOINT, CAM_LOOKAT, CAM_UP);
		break;
	case 'i':
		showInfo = !showInfo;
		break;
	case 'o':
		TakeScreenShot();
		break;
	case '1':
		pPSystem->Reset();
		pPSystem = &gFountain;
		break;
	case '2':
		pPSystem->Reset();
		pPSystem = &gSmokeCloud;
		break;
	case '3':
		pPSystem->Reset();
		pPSystem = &gSmokeStack;
		break;
	};
}

//------------------------------------------------------
void glutMouseCallback(int button, int state, int x, int y)
{
	if(state == GLUT_DOWN)
	{
		lastMouseX = x;
		lastMouseY = y;

		mouseZoom = false;
		mouseMove = false;

		if(button == GLUT_RIGHT_BUTTON)
			mouseZoom = true;
		else if(button == GLUT_MIDDLE_BUTTON)
			mouseMove = true;
	}
}

//------------------------------------------------------
void glutMouseMoveCallback(int x, int y)
{
	float dx = (x - lastMouseX) / (float)WIDTH;
	float dy = (y - lastMouseY) / (float)HEIGHT;

	if(mouseZoom)
	{
		gCam.Zoom(-100.0f * dy);
	}
	else if(mouseMove)
	{
		gCam.Move( Vector3f(10.0f * dx, 0, -10.0f * dy) );
	}
	else
	{
		gCam.RotateRight(4.0f * dx);
		gCam.RotateUp(-4.0f * dy);
	}

	lastMouseX = x;
	lastMouseY = y;
}

//------------------------------------------------------
bool InitProgram(int w, int h)
{
	glClearColor(bgColor.r, bgColor.g, bgColor.b, bgColor.a);

	glMatrixMode(GL_PROJECTION);
	glLoadIdentity();

	// Sets matrix projection
	gProj.MakePerspective((float)w, (float)h, 1.0f, 100.0f);

	float matProj[16];
	gProj.GetOpenGLMatrix(matProj);
	glMultMatrixf(matProj);

	glEnable(GL_DEPTH_TEST);
	glEnable(GL_TEXTURE_2D);
	glEnable(GL_BLEND); 
	glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);

	gCam.Set(CAM_EYEPOINT, CAM_LOOKAT, CAM_UP);

	// Loads texture data
	floorCenterTexID = SOIL_load_OGL_texture(FLOOR_CENTER_IMAGE_FILE.c_str(),
											 SOIL_LOAD_RGB,
											 SOIL_CREATE_NEW_ID,
											 SOIL_FLAG_INVERT_Y);

	floorEdgeTexID = SOIL_load_OGL_texture(FLOOR_EDGE_IMAGE_FILE.c_str(),
										   SOIL_LOAD_RGB,
										   SOIL_CREATE_NEW_ID,
										   SOIL_FLAG_INVERT_Y);

	particleTexID = SOIL_load_OGL_texture(PARTICLE_FOUNTAIN_IMAGE_FILE.c_str(),
                                          SOIL_LOAD_RGBA,
									      SOIL_CREATE_NEW_ID,
									      SOIL_FLAG_INVERT_Y);

	smokeTexID = SOIL_load_OGL_texture(PARTICLE_SMOKE_IMAGE_FILE.c_str(),
                                       SOIL_LOAD_RGBA,
									   SOIL_CREATE_NEW_ID,
									   SOIL_FLAG_INVERT_Y);

	// Sets up fountain particle system
	gFountain.SetTexture(particleTexID);

	Plane pl;
	pl.n = Vector3f(0.0f, 1.0f, 0.0f);
	pl.d = 0.0f;

	gFountain.AddPlane(pl);
	gFountain.SetBounce(0.5f);
	gFountain.SetSpawnRate(50.0f);
	gFountain.SetMaxParticles(1000);
	gFountain.SetParticleLife(3.0f);
	gFountain.SetStartPosition(Point3f(0.0f, 0.0f, 0.0f));
	gFountain.SetGravity(Vector3f(0.0f, -9.8f, 0.0f));
	gFountain.SetAlphaFunc(AlphaFuncPoly4);

	gFountain.Init();

	// Sets up smoke cloud particle system
	gSmokeCloud.SetTexture(smokeTexID);
	gSmokeCloud.SetSpawnRate(50.0f);
	gSmokeCloud.SetMaxParticles(1000);
	gSmokeCloud.SetParticleLife(5.0f);
	gSmokeCloud.SetStartPosition(Point3f(0.0f, 1.0f, 0.0f));
	gSmokeCloud.SetGravity(Vector3f(0.0f, 0.0f, 0.0f));
	gSmokeCloud.SetInitFunc(SmokeCloudParticleInit);
	gSmokeCloud.SetAlphaFunc(AlphaFuncLinear);
	gSmokeCloud.SetColorFunc(SmokeStackColorFunc);

	gSmokeCloud.Init();

	// Sets up the smoke stack particle system
	gSmokeStack.SetTexture(smokeTexID);
	gSmokeStack.SetSpawnRate(50.0f);
	gSmokeStack.SetMaxParticles(1000);
	gSmokeStack.SetParticleLife(5.0f);
	gSmokeStack.SetStartPosition(Point3f(0.0f, 0.0f, 0.0f));
	gSmokeStack.SetGravity(Vector3f(1.0f, 5.0f, 0.0f));
	gSmokeStack.SetInitFunc(SmokeStackParticleInit);
	gSmokeStack.SetAlphaFunc(AlphaFuncPoly3);
	gSmokeStack.SetColorFunc(SmokeStackColorFunc);
	//gSmokeStack.SetSizeFunc(SmokeStackSizeFunc);

	gSmokeStack.Init();

	pPSystem = &gFountain;

	return InitFBO() && InitShaders();
}

//------------------------------------------------------
bool ReInitProgram(int w, int h)
{
	WIDTH = w;
	HEIGHT = h;

	glMatrixMode(GL_PROJECTION);
	glLoadIdentity();

	// Sets matrix projection
	gProj.MakePerspective((float)w, (float)h, 1.0f, 1000.0f);

	glViewport(0, 0, w, h);

	float matProj[16];
	gProj.GetOpenGLMatrix(matProj);
	glMultMatrixf(matProj);

	return InitFBO();
}

//------------------------------------------------------
bool InitFBO()
{
	if(fbo)
	{
		// Cleans up frame buffer object
		glDeleteFramebuffersEXT(1, (GLuint*)&fbo);
		
		// Cleans up depth
		glDeleteTextures(1, (GLuint*)&depthTex);
	}

	glGenFramebuffersEXT(1, (GLuint*)&fbo);
	glBindFramebufferEXT(GL_FRAMEBUFFER_EXT, fbo);

	glGenTextures(1, (GLuint*)&depthTex);
	glBindTexture(GL_TEXTURE_2D, depthTex);
	glTexImage2D(GL_TEXTURE_2D, 0, GL_RGB32F_ARB, WIDTH, HEIGHT, 0, GL_RGB, GL_FLOAT, NULL);
	glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_CLAMP_TO_EDGE);
	glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_CLAMP_TO_EDGE);
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR);
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR);

	glFramebufferTexture2DEXT(GL_FRAMEBUFFER_EXT, GL_COLOR_ATTACHMENT0_EXT, GL_TEXTURE_2D, depthTex, 0);

	GLuint err = glCheckFramebufferStatusEXT(GL_FRAMEBUFFER_EXT);
	if(err != GL_FRAMEBUFFER_COMPLETE_EXT)
	{
		std::cerr << "ERROR: frame buffer object could not be created.\n";
		return false;
	}

	// Unbinds the fbo until we need it
	glBindFramebufferEXT(GL_FRAMEBUFFER_EXT, 0);

	return true;
}

//------------------------------------------------------
bool LoadShaderSource(const char *name, std::string &data)
{
	// open the shader file
	std::ifstream file(name);
	if(!file.good()) 
		return false;
	
	// length of the shader
	file.seekg(0, std::ios_base::end);
	int length = file.tellg();
	length += 1; 

	// read the source
	data.resize(length);
	file.seekg(0, std::ios_base::beg);
	file.read(&data[0], length);
	data[length-1] = '\0';
	
	return true;
}

//------------------------------------------------------
void PrintShaderInfo(const char *name, GLhandleARB shader)
{
	int infologLength = 0;
	glGetObjectParameterivARB(shader, GL_OBJECT_INFO_LOG_LENGTH_ARB, (GLint*)&infologLength);

	if(infologLength > 0)
    {
        int charsWritten;
		GLcharARB *infoLog = new GLcharARB[infologLength];
        glGetInfoLogARB(shader, infologLength, (GLsizei*)&charsWritten, infoLog);

		std::cout << name << " info:\n" << infoLog << std::endl;
        
		delete [] infoLog;
    }
}

//------------------------------------------------------
bool LoadShader(const char *vs, const char *ps, GLhandleARB &vsID, GLhandleARB &psID, GLhandleARB &prog)
{
	std::cout << "----------------------------------------------------------------" << std::endl;

	// load shader source
	std::string vert, frag;
	bool vertFlag = LoadShaderSource(vs, vert);
	bool fragFlag = LoadShaderSource(ps, frag);
	if( !vertFlag || !fragFlag ) 
	{
		std::cerr << "ERROR: Cannot load vertex/fragment program ... " << std::endl;
		return false;
	}

	// Load source code strings into shaders
	vsID = glCreateShaderObjectARB(GL_VERTEX_SHADER_ARB);
	const GLcharARB *source1 = (GLcharARB*)vert.c_str();
    glShaderSourceARB(vsID, 1, &source1, NULL);

	psID = glCreateShaderObjectARB(GL_FRAGMENT_SHADER_ARB);
	const GLcharARB *source2 = (GLcharARB*)frag.c_str();
    glShaderSourceARB(psID, 1, &source2, NULL);

	// compile the vertex and fragment program
	GLint vertCompiled, fragCompiled;
	
	glCompileShaderARB(vsID);
	glGetObjectParameterivARB(vsID, GL_OBJECT_COMPILE_STATUS_ARB, &vertCompiled);
	PrintShaderInfo(vs, vsID);

	glCompileShaderARB(psID);
	glGetObjectParameterivARB(psID, GL_OBJECT_COMPILE_STATUS_ARB, &fragCompiled);
	PrintShaderInfo(ps, psID);

	 if(!vertCompiled || !fragCompiled) 
		 return false;

	// Link the program
	GLint linked;
    prog = glCreateProgramObjectARB();
	
	glAttachObjectARB(prog, vsID);
	glAttachObjectARB(prog, psID);
	glLinkProgramARB(prog);

	// linked status
	glGetObjectParameterivARB(prog, GL_OBJECT_LINK_STATUS_ARB, &linked);
	PrintShaderInfo("Shader", prog);

	std::cout << "----------------------------------------------------------------" << std::endl;

	return true;
}

//------------------------------------------------------
bool InitShaders()
{
	bool r1 = LoadShader(DEPTH_OUT_VS_FILE.c_str(), 
                         DEPTH_OUT_PS_FILE.c_str(), 
						 depthVS, 
						 depthPS, 
						 depthShader);

	bool r2 = LoadShader(PARTICLE_SYSTEM_VS_FILE.c_str(), 
                         PARTICLE_SYSTEM_PS_FILE.c_str(), 
						 particleVS, 
						 particlePS, 
						 particleShader);

	return r1 && r2;
}

//------------------------------------------------------
void DrawFloorQuad(float offx, float offz)
{
	float w = 5.0f;
	float d = 5.0f;

	offx = 2.0f * w * offx; 
	offz = 2.0f * d * offz;

	glTexCoord2f(0.0f, 0.0f);
	glVertex3f(-w + offx, 0.0f, -d + offz);

	glTexCoord2f(1.0f, 0.0f);
	glVertex3f( w + offx, 0.0f, -d + offz);

	glTexCoord2f(1.0f, 1.0f);
	glVertex3f( w + offx, 0.0f,  d + offz);

	glTexCoord2f(0.0f, 1.0f);
	glVertex3f(-w + offx, 0.0f,  d + offz);
}

//------------------------------------------------------
void DrawSceneObjects()
{
	glMatrixMode(GL_MODELVIEW);
	glPushMatrix();

	// Draws center floor quad
	glBindTexture(GL_TEXTURE_2D, floorCenterTexID);

	glBegin(GL_QUADS);

		DrawFloorQuad( 0.0f,  0.0f);

	glEnd();

	// Draws edge quads
	glBindTexture(GL_TEXTURE_2D, floorEdgeTexID);

	glBegin(GL_QUADS);

		DrawFloorQuad( 0.0f,  1.0f);
		DrawFloorQuad( 0.0f, -1.0f);
		DrawFloorQuad( 1.0f,  0.0f);
		DrawFloorQuad( 1.0f,  1.0f);
		DrawFloorQuad( 1.0f, -1.0f);
		DrawFloorQuad(-1.0f,  0.0f);
		DrawFloorQuad(-1.0f,  1.0f);
		DrawFloorQuad(-1.0f, -1.0f);

	glEnd();

	glBindTexture(GL_TEXTURE_2D, 0);
	
	glPopMatrix();
}

//------------------------------------------------------
void RenderScene()
{
	glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);

	// Sets view matrix
	glMatrixMode(GL_MODELVIEW);
	glLoadIdentity();

	Matrixf viewTrans = gCam.GetViewMatrix();

	float matView[16];
	viewTrans.GetOpenGLMatrix(matView);
	glMultMatrixf(matView);

	if(useSoftParticles)
	{
		// Renders depth information
		glBindFramebufferEXT(GL_FRAMEBUFFER_EXT, fbo);

		glClampColorARB(GL_CLAMP_VERTEX_COLOR_ARB, GL_FALSE);
		glClampColorARB(GL_CLAMP_FRAGMENT_COLOR_ARB, GL_FALSE);
		glClampColorARB(GL_CLAMP_READ_COLOR_ARB, GL_FALSE);

		glClearColor(FLT_MAX, FLT_MAX, FLT_MAX, FLT_MAX);
		glClear(GL_COLOR_BUFFER_BIT);
		glClearColor(bgColor.r, bgColor.g, bgColor.b, bgColor.a);

		// Renders depth information of the scene
		glUseProgramObjectARB(depthShader);
	
		DrawSceneObjects();
		
		glUseProgramObjectARB(0);
		glBindFramebufferEXT(GL_FRAMEBUFFER_EXT, 0);

		glClampColorARB(GL_CLAMP_VERTEX_COLOR_ARB, GL_TRUE);
		glClampColorARB(GL_CLAMP_FRAGMENT_COLOR_ARB, GL_TRUE);
		glClampColorARB(GL_CLAMP_READ_COLOR_ARB, GL_TRUE);

		glBindTexture(GL_TEXTURE_2D, 0); 

		// Draws the scene normally
		DrawSceneObjects();

		// renders the soft particles
		glUseProgramObjectARB(particleShader);

		// Sets texture data
		GLint texloc = glGetUniformLocationARB(particleShader, "tex");
		GLint depthTexloc = glGetUniformLocationARB(particleShader, "depthInfo");
		GLint powerloc = glGetUniformLocationARB(particleShader, "power");

		glUniform1fARB(powerloc, (float)softParticlePower);

		glActiveTexture(GL_TEXTURE0);
		glBindTexture(GL_TEXTURE_2D, pPSystem->TextureID());
		glUniform1iARB(texloc, 0);

		glActiveTexture(GL_TEXTURE1);
		glBindTexture(GL_TEXTURE_2D, depthTex);
		glUniform1iARB(depthTexloc, 1);

		pPSystem->Draw(gCam, false);

		// Unbinds shader and textures
		glBindTexture(GL_TEXTURE_2D, 0);
		glActiveTexture(GL_TEXTURE0);
		glBindTexture(GL_TEXTURE_2D, 0);
		glUseProgramObjectARB(0);
	}
	else
	{
		DrawSceneObjects();

		// Draws particle system
		pPSystem->Draw(gCam, true);
	};

	if(showInfo)
	{
		DrawFPS();
		DrawSceneInfo();
	}

	glutSwapBuffers();
}

//------------------------------------------------------
void DrawFPS()
{
	std::stringstream str;
	str << "FPS: " << fps;

	glutDrawStr(Colors::Red, 
                -WIDTH / 2.0f + 5.0f, 
				HEIGHT / 2.0f - 20.0f, 
				str.str().c_str());
}

//------------------------------------------------------
void DrawSceneInfo()
{
	if(pausePSystem)
	{
		glutDrawStr(Colors::Red, 
					WIDTH / 2.0f - 80.0f, 
					HEIGHT / 2.0f - 20.0f, 
					"PAUSED");
	}

	std::stringstream sp;
	sp << "Soft Particles: ";
	if(softParticleSupport)
	{
		if(useSoftParticles)
			sp << "ON";
		else
			sp << "OFF";
	}
	else
	{
		sp << "NO HARDWARE SUPPORT";
	}
	glutDrawStr(Colors::White,
                -WIDTH / 2.0f + 5.0f, 
				-HEIGHT / 2.0f + 5.0f,
				sp.str().c_str());

	std::stringstream power;
	power << "Soft Particle Power: " << (float)softParticlePower;
	glutDrawStr(Colors::White,
                -WIDTH / 2.0f + 5.0f, 
				-HEIGHT / 2.0f + 30.0f,
				power.str().c_str());

	std::stringstream wind;
	wind << "Wind: ";
	if(pPSystem->GetWind() == WIND)
		wind << "ON";
	else
		wind << "OFF";
	glutDrawStr(Colors::White,
                -WIDTH / 2.0f + 5.0f, 
				-HEIGHT / 2.0f + 55.0f,
				wind.str().c_str());

	std::stringstream spawn;
	spawn << "Spawn Rate: " << pPSystem->GetSpawnRate();
	glutDrawStr(Colors::White,
                -WIDTH / 2.0f + 5.0f, 
				-HEIGHT / 2.0f + 80.0f,
				spawn.str().c_str());

	std::stringstream particles;
	particles << "Particles: " << pPSystem->ParticleCount();
	glutDrawStr(Colors::White,
                -WIDTH / 2.0f + 5.0f, 
				-HEIGHT / 2.0f + 102.0f,
				particles.str().c_str());
}

//------------------------------------------------------
void glutDrawStr(const Color4f &c, float x, float y, const char *str)
{
	glMatrixMode(GL_PROJECTION);
	glPushMatrix();
	glLoadIdentity();
	
	float w = WIDTH / 2.0f;
	float h = HEIGHT / 2.0f;
	glOrtho(-w, w, -h, h, -100.0f, 100.0f);

	glMatrixMode(GL_MODELVIEW);
	glPushMatrix();
	glLoadIdentity();

	glTranslatef(x, y, 0.0f);
	glScalef(0.15f, 0.15f, 0.15f);

	glColor4f(c.r, c.g, c.b, c.a);

	for(const char *p = str; *p; p++)
	{
		glutStrokeCharacter(GLUT_STROKE_ROMAN, *p);
	}

	glColor4f(1.0f, 1.0f, 1.0f, 1.0f);

	glPopMatrix();

	glMatrixMode(GL_PROJECTION);
	glPopMatrix();
}

//------------------------------------------------------
void TakeScreenShot()
{
	time_t rawTime = time(NULL);
	tm *timeInfo = localtime(&rawTime);
	
	char timeStr[64];
	strftime(timeStr, 64, "%m%d%Y_%H%M%S", timeInfo);
	
	std::stringstream file;
	file << "Screenshot_" << timeStr << ".bmp";
	
	SOIL_save_screenshot(file.str().c_str(), SOIL_SAVE_TYPE_BMP, 0, 0, WIDTH, HEIGHT);
}

//------------------------------------------------------
void CleanUp()
{
	// Cleans up frame buffer object
	glDeleteFramebuffersEXT(1, &fbo);

	// Cleans up shaders
	glDeleteObjectARB(depthVS);
	glDeleteObjectARB(depthPS);
	glDeleteObjectARB(depthShader);
	glDeleteObjectARB(particleVS);
	glDeleteObjectARB(particlePS);
	glDeleteObjectARB(particleShader);
	
	// Cleans up textures
	glDeleteTextures(1, &depthTex);
	glDeleteTextures(1, &floorCenterTexID);\
	glDeleteTextures(1, &floorEdgeTexID);
	glDeleteTextures(1, &particleTexID);
	glDeleteTextures(1, &smokeTexID);
}
